/* 
 * tclTestObj.c --
 *
 *	This file contains C command procedures for the additional Tcl
 *	commands that are used for testing implementations of the Tcl object
 *	types. These commands are not normally included in Tcl
 *	applications; they're only used for testing.
 *
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclTestObj.c 1.15 97/01/21 21:30:17
 */

#include "tclInt.h"

/*
 * An array of Tcl_Obj pointers used in the commands that operate on or get
 * the values of Tcl object-valued variables. varPtr[i] is the i-th
 * variable's Tcl_Obj *.
 */

#define NUMBER_OF_OBJECT_VARS 20
static Tcl_Obj *varPtr[NUMBER_OF_OBJECT_VARS];

/*
 * Forward declarations for procedures defined later in this file:
 */

static int		CheckIfVarUnset _ANSI_ARGS_((Tcl_Interp *interp,
			    int varIndex));
static int		GetVariableIndex _ANSI_ARGS_((Tcl_Interp *interp,
			    char *string, int *indexPtr));
static void		SetVarToExistingObj _ANSI_ARGS_((int varIndex,
			    Tcl_Obj *objPtr));
static void		SetVarToNewObj _ANSI_ARGS_((int varIndex,
			    Tcl_Obj *objPtr));
int			TclObjTest_Init _ANSI_ARGS_((Tcl_Interp *interp));
static int		TestbooleanobjCmd _ANSI_ARGS_((ClientData dummy,
			    Tcl_Interp *interp, int objc, Tcl_Obj **objv));
static int		TestconvertobjCmd _ANSI_ARGS_((ClientData dummy,
			    Tcl_Interp *interp, int objc, Tcl_Obj **objv));
static int		TestdoubleobjCmd _ANSI_ARGS_((ClientData dummy,
			    Tcl_Interp *interp, int objc, Tcl_Obj **objv));
static int		TestindexobjCmd _ANSI_ARGS_((ClientData dummy,
			    Tcl_Interp *interp, int objc, Tcl_Obj **objv));
static int		TestintobjCmd _ANSI_ARGS_((ClientData dummy,
			    Tcl_Interp *interp, int objc, Tcl_Obj **objv));
static int		TestobjCmd _ANSI_ARGS_((ClientData dummy,
			    Tcl_Interp *interp, int objc, Tcl_Obj **objv));
static int		TeststringobjCmd _ANSI_ARGS_((ClientData dummy,
			    Tcl_Interp *interp, int objc, Tcl_Obj **objv));

/*
 *----------------------------------------------------------------------
 *
 * TclObjTest_Init --
 *
 *	This procedure creates additional commands that are used to test the
 *	Tcl object support.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Creates and registers several new testing commands.
 *
 *----------------------------------------------------------------------
 */

int
TclObjTest_Init(interp)
    Tcl_Interp *interp;
{
    register int i;
    
    for (i = 0;  i < NUMBER_OF_OBJECT_VARS;  i++) {
        varPtr[i] = NULL;
    }
	
    Tcl_CreateObjCommand(interp, "testbooleanobj", -1, TestbooleanobjCmd,
	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "testconvertobj", -1, TestconvertobjCmd,
	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "testdoubleobj", -1, TestdoubleobjCmd,
	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "testintobj", -1, TestintobjCmd,
	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "testindexobj", -1, TestindexobjCmd,
	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "testobj", -1, TestobjCmd,
	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "teststringobj", -1, TeststringobjCmd,
	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TestbooleanobjCmd --
 *
 *	This procedure implements the "testbooleanobj" command.  It is used
 *	to test the boolean Tcl object type implementation.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	Creates and frees boolean objects, and also converts objects to
 *	have boolean type.
 *
 *----------------------------------------------------------------------
 */

static int
TestbooleanobjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    int varIndex, boolValue, length;
    Tcl_Obj *newObjPtr;
    char *index, *subCmd, *string;
	
    if (objc < 3) {
	wrongNumArgs:
	TclSetObjResultFromString(interp, "wrong # args", -1);
	return TCL_ERROR;
    }

    /*
     * THIS FAILS IF AN OBJECT'S STRING REP HAS A NULL BYTE.
     */

    index = Tcl_GetStringFromObj(objv[2], &length);
    if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
	return TCL_ERROR;
    }

    subCmd = Tcl_GetStringFromObj(objv[1], &length);
    if (strcmp(subCmd, "set") == 0) {
	if (objc != 4) {
	    goto wrongNumArgs;
	}
	if (Tcl_GetBooleanFromObj(interp, objv[3], &boolValue) != TCL_OK) {
	    TclSetObjResultFromStrResult(interp);
	    return TCL_ERROR;
	}

	/*
	 * If the object currently bound to the variable with index varIndex
	 * has ref count 1 (i.e. the object is unshared) we can modify that
	 * object directly. Otherwise, if RC>1 (i.e. the object is shared),
	 * we must create a new object to modify/set and decrement the old
	 * formerly-shared object's ref count. This is "copy on write".
	 */

	if ((varPtr[varIndex] != NULL) && !Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetBooleanObj(varPtr[varIndex], boolValue);
	} else {
	    newObjPtr = Tcl_NewBooleanObj(boolValue);
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "get") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "not") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	if (Tcl_GetBooleanFromObj(interp, varPtr[varIndex],
				  &boolValue) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (!Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetBooleanObj(varPtr[varIndex], !boolValue);
	} else {
	    newObjPtr = Tcl_NewBooleanObj(!boolValue);
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else {
	Tcl_Obj *resultPtr;
	
	Tcl_ResetObjResult(interp);
        resultPtr = ((Interp *) interp)->objResult;
        Tcl_StringObjAppend(resultPtr, "bad option \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[1]);
        Tcl_StringObjAppend(resultPtr,
                 "\": must be set, get, or not", -1);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TestconvertobjCmd --
 *
 *	This procedure implements the "testconvertobj" command. It is used
 *	to test converting objects to new types.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	Converts objects to new types.
 *
 *----------------------------------------------------------------------
 */

static int
TestconvertobjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    int length;
    char *subCmd;
    char buf[20];
    Tcl_Obj *resultPtr;
	
    if (objc < 3) {
	wrongNumArgs:
	TclSetObjResultFromString(interp, "wrong # args", -1);
	return TCL_ERROR;
    }

    /*
     * THIS FAILS IF AN OBJECT'S STRING REP HAS A NULL BYTE.
     */

    subCmd = Tcl_GetStringFromObj(objv[1], &length);
    if (strcmp(subCmd, "double") == 0) {
	double d;

	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (Tcl_GetDoubleFromObj(interp, objv[2], &d) != TCL_OK) {
	    return TCL_ERROR;
	}
	sprintf(buf, "%f", d);
        TclSetObjResultFromString(interp, buf, -1);
    } else {
	Tcl_ResetObjResult(interp);
        resultPtr = ((Interp *) interp)->objResult;
        Tcl_StringObjAppend(resultPtr, "bad option \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[1]);
        Tcl_StringObjAppend(resultPtr, "\": must be double", -1);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TestdoubleobjCmd --
 *
 *	This procedure implements the "testdoubleobj" command.  It is used
 *	to test the double-precision floating point Tcl object type
 *	implementation.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	Creates and frees double objects, and also converts objects to
 *	have double type.
 *
 *----------------------------------------------------------------------
 */

static int
TestdoubleobjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    int varIndex, length;
    double doubleValue;
    Tcl_Obj *newObjPtr;
    char *index, *subCmd, *string;
	
    if (objc < 3) {
	wrongNumArgs:
	TclSetObjResultFromString(interp, "wrong # args", -1);
	return TCL_ERROR;
    }

    /*
     * THIS FAILS IF AN OBJECT'S STRING REP HAS A NULL BYTE.
     */

    index = Tcl_GetStringFromObj(objv[2], &length);
    if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
	return TCL_ERROR;
    }

    subCmd = Tcl_GetStringFromObj(objv[1], &length);
    if (strcmp(subCmd, "set") == 0) {
	if (objc != 4) {
	    goto wrongNumArgs;
	}
	string = Tcl_GetStringFromObj(objv[3], &length);
	if (Tcl_GetDouble(interp, string, &doubleValue) != TCL_OK) {
	    TclSetObjResultFromStrResult(interp);
	    return TCL_ERROR;
	}

	/*
	 * If the object currently bound to the variable with index varIndex
	 * has ref count 1 (i.e. the object is unshared) we can modify that
	 * object directly. Otherwise, if RC>1 (i.e. the object is shared),
	 * we must create a new object to modify/set and decrement the old
	 * formerly-shared object's ref count. This is "copy on write".
	 */

	if ((varPtr[varIndex] != NULL) && !Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetDoubleObj(varPtr[varIndex], doubleValue);
	} else {
	    newObjPtr = Tcl_NewDoubleObj(doubleValue);
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "get") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "mult10") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	if (Tcl_GetDoubleFromObj(interp, varPtr[varIndex],
				 &doubleValue) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (!Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetDoubleObj(varPtr[varIndex], (doubleValue * 10.0));
	} else {
	    newObjPtr = Tcl_NewDoubleObj( (doubleValue * 10.0) );
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "div10") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	if (Tcl_GetDoubleFromObj(interp, varPtr[varIndex],
				 &doubleValue) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (!Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetDoubleObj(varPtr[varIndex], (doubleValue / 10.0));
	} else {
	    newObjPtr = Tcl_NewDoubleObj( (doubleValue / 10.0) );
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else {
	Tcl_Obj *resultPtr;
	
	Tcl_ResetObjResult(interp);
        resultPtr = ((Interp *) interp)->objResult;
        Tcl_StringObjAppend(resultPtr, "bad option \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[1]);
        Tcl_StringObjAppend(resultPtr,
                 "\": must be set, get, mult10, or div10", -1);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TestindexobjCmd --
 *
 *	This procedure implements the "testindexobj" command. It is used to
 *	test the index Tcl object type implementation.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	Creates and frees int objects, and also converts objects to
 *	have int type.
 *
 *----------------------------------------------------------------------
 */

static int
TestindexobjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    int index, setError, i, dummy, result;
    long index2;
    char **argv;
    static char *tablePtr[] = {"a", "b", "check", (char *) NULL};

    if ((objc == 3) && (strcmp(Tcl_GetStringFromObj(objv[1], &dummy),
	    "check") == 0)) {
	/*
	 * This code checks to be sure that the results of
	 * Tcl_GetIndexFromObj are properly cached in the object and
	 * returned on subsequent lookups.
	 */

	Tcl_GetIndexFromObj((Tcl_Interp *) NULL, objv[1], tablePtr,
		"token", &index);
	if (Tcl_GetIntFromObj(interp, objv[2], &index2) != TCL_OK) {
	    return TCL_ERROR;
	}
	objv[1]->internalRep.twoPtrValue.ptr2 = (VOID *) index2;
	result = Tcl_GetIndexFromObj((Tcl_Interp *) NULL, objv[1],
		tablePtr, "token", &index);
	if (result == TCL_OK) {
	    Tcl_SetIntObj(Tcl_GetObjResult(interp), index);
	}
	return result;
    }

    if (objc < 4) {
	TclSetObjResultFromString(interp, "wrong # args", -1);
	return TCL_ERROR;
    }

    if (Tcl_GetBooleanFromObj(interp, objv[1], &setError) != TCL_OK) {
	return TCL_ERROR;
    }
    argv = (char **) ckalloc((unsigned) ((objc-2) * sizeof(char *)));
    for (i = 3; i < objc; i++) {
	argv[i-3] = Tcl_GetStringFromObj(objv[i], &dummy);
    }
    argv[objc-3] = NULL;
    result = Tcl_GetIndexFromObj(setError ? interp : NULL, objv[2],
	    argv, "token", &index);
    ckfree((char *) argv);
    if (result == TCL_OK) {
	Tcl_SetIntObj(Tcl_GetObjResult(interp), index);
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * TestintobjCmd --
 *
 *	This procedure implements the "testintobj" command. It is used to
 *	test the int Tcl object type implementation.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	Creates and frees int objects, and also converts objects to
 *	have int type.
 *
 *----------------------------------------------------------------------
 */

static int
TestintobjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    long intValue;
    int varIndex, length;
    Tcl_Obj *newObjPtr;
    char *index, *subCmd, *string;
	
    if (objc < 3) {
	wrongNumArgs:
	TclSetObjResultFromString(interp, "wrong # args", -1);
	return TCL_ERROR;
    }

    /*
     * THIS FAILS IF AN OBJECT'S STRING REP HAS A NULL BYTE.
     */

    index = Tcl_GetStringFromObj(objv[2], &length);
    if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
	return TCL_ERROR;
    }

    subCmd = Tcl_GetStringFromObj(objv[1], &length);
    if (strcmp(subCmd, "set") == 0) {
	int i;
	if (objc != 4) {
	    goto wrongNumArgs;
	}
	string = Tcl_GetStringFromObj(objv[3], &length);
	if (Tcl_GetInt(interp, string, &i) != TCL_OK) {
	    TclSetObjResultFromStrResult(interp);
	    return TCL_ERROR;
	}
	intValue = i;

	/*
	 * If the object currently bound to the variable with index varIndex
	 * has ref count 1 (i.e. the object is unshared) we can modify that
	 * object directly. Otherwise, if RC>1 (i.e. the object is shared),
	 * we must create a new object to modify/set and decrement the old
	 * formerly-shared object's ref count. This is "copy on write".
	 */

	if ((varPtr[varIndex] != NULL) && !Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetIntObj(varPtr[varIndex], intValue);
	} else {
	    newObjPtr = Tcl_NewIntObj(intValue);
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "get") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "mult10") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	if (Tcl_GetIntFromObj(interp, varPtr[varIndex],
			      &intValue) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (!Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetIntObj(varPtr[varIndex], (intValue * 10));
	} else {
	    newObjPtr = Tcl_NewIntObj( (intValue * 10) );
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "div10") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	if (Tcl_GetIntFromObj(interp, varPtr[varIndex],
			      &intValue) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (!Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetIntObj(varPtr[varIndex], (intValue / 10));
	} else {
	    newObjPtr = Tcl_NewIntObj( (intValue / 10) );
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else {
	Tcl_Obj *resultPtr;
	
	Tcl_ResetObjResult(interp);
        resultPtr = ((Interp *) interp)->objResult;
        Tcl_StringObjAppend(resultPtr, "bad option \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[1]);
        Tcl_StringObjAppend(resultPtr,
                 "\": must be set, get, mult10, or div10", -1);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TestobjCmd --
 *
 *	This procedure implements the "testobj" command. It is used to test
 *	the type-independent portions of the Tcl object type implementation.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	Creates and frees objects.
 *
 *----------------------------------------------------------------------
 */

static int
TestobjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    int varIndex, destIndex, i;
    Tcl_Obj *typeListPtr, *newObjPtr;
    char *index, *subCmd, *string;
    Tcl_ObjType *targetType;
    char buf[20];
    int length;
	
    if (objc < 2) {
	wrongNumArgs:
	TclSetObjResultFromString(interp, "wrong # args", -1);
	return TCL_ERROR;
    }

    /*
     * THIS FAILS IF AN OBJECT'S STRING REP HAS A NULL BYTE.
     */

    subCmd = Tcl_GetStringFromObj(objv[1], &length);
    if (strcmp(subCmd, "assign") == 0) {
        if (objc != 4) {
            goto wrongNumArgs;
        }
        index = Tcl_GetStringFromObj(objv[2], &length);
        if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	string = Tcl_GetStringFromObj(objv[3], &length);
        if (GetVariableIndex(interp, string, &destIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        SetVarToExistingObj(destIndex, varPtr[varIndex]);
        string = Tcl_GetStringFromObj(varPtr[destIndex], &length);
        TclSetObjResultFromString(interp, string, length);
     } else if (strcmp(subCmd, "convert") == 0) {
        char *typeName;
        if (objc != 4) {
            goto wrongNumArgs;
        }
        index = Tcl_GetStringFromObj(objv[2], &length);
        if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
        typeName = Tcl_GetStringFromObj(objv[3], &length);
        if ((targetType = Tcl_GetObjType(typeName)) == NULL) {
	    Tcl_Obj *resultPtr;
	    
	    Tcl_ResetObjResult(interp);
	    resultPtr = ((Interp *) interp)->objResult;
	    Tcl_StringObjAppend(resultPtr, "no type ", -1);
	    Tcl_StringObjAppend(resultPtr, typeName, -1);
	    Tcl_StringObjAppend(resultPtr, " found", -1);
            return TCL_ERROR;
        }
        if (Tcl_ConvertToType(interp, varPtr[varIndex], targetType)
            != TCL_OK) {
            return TCL_ERROR;
        }
        string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
        TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "duplicate") == 0) {
        if (objc != 4) {
            goto wrongNumArgs;
        }
        index = Tcl_GetStringFromObj(objv[2], &length);
        if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	string = Tcl_GetStringFromObj(objv[3], &length);
        if (GetVariableIndex(interp, string, &destIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        newObjPtr = Tcl_DuplicateObj(varPtr[varIndex]);
        SetVarToNewObj(destIndex, newObjPtr);
        string = Tcl_GetStringFromObj(varPtr[destIndex], &length);
        TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "freeallvars") == 0) {
        if (objc != 2) {
            goto wrongNumArgs;
        }
        for (i = 0;  i < NUMBER_OF_OBJECT_VARS;  i++) {
            if (varPtr[i] != NULL) {
                Tcl_DecrRefCount(varPtr[i]);
                varPtr[i] = NULL;
            }
        }
    } else if (strcmp(subCmd, "newobj") == 0) {
        if (objc != 3) {
            goto wrongNumArgs;
        }
        index = Tcl_GetStringFromObj(objv[2], &length);
        if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        SetVarToNewObj(varIndex, Tcl_NewObj());
        string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
        TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "refcount") == 0) {
        if (objc != 3) {
            goto wrongNumArgs;
        }
        index = Tcl_GetStringFromObj(objv[2], &length);
        if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
        sprintf(buf, "%d", varPtr[varIndex]->refCount);
        TclSetObjResultFromString(interp, buf, -1);
    } else if (strcmp(subCmd, "type") == 0) {
        if (objc != 3) {
            goto wrongNumArgs;
        }
        index = Tcl_GetStringFromObj(objv[2], &length);
        if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
            return TCL_ERROR;
        }
        if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
        if (varPtr[varIndex]->typePtr == NULL) { /* a string! */
	    TclSetObjResultFromString(interp, "string", -1);
        } else {
            TclSetObjResultFromString(interp,
                                      varPtr[varIndex]->typePtr->name, -1);
        }
    } else if (strcmp(subCmd, "types") == 0) {
        if (objc != 2) {
            goto wrongNumArgs;
        }
        if (Tcl_GetAllObjTypes(interp, &typeListPtr) != TCL_OK) {
            return TCL_ERROR;
        }
        string = Tcl_GetStringFromObj(typeListPtr, &length);
        TclSetObjResultFromString(interp, string, length);
        Tcl_DecrRefCount(typeListPtr);
    } else {
	Tcl_Obj *resultPtr;
	
	Tcl_ResetObjResult(interp);
        resultPtr = ((Interp *) interp)->objResult;
        Tcl_StringObjAppend(resultPtr, "bad option \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[1]);
        Tcl_StringObjAppend(resultPtr,
                 "\": must be assign, convert, duplicate, freeallvars, newobj, objcount, refcount, type, or types", -1);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TeststringobjCmd --
 *
 *	This procedure implements the "teststringobj" command. It is used to
 *	test the string Tcl object type implementation.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	Creates and frees string objects, and also converts objects to
 *	have string type.
 *
 *----------------------------------------------------------------------
 */

static int
TeststringobjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    int varIndex, length;
    Tcl_Obj *newObjPtr;
    char *index, *subCmd, *string;
	
    if (objc < 3) {
	wrongNumArgs:
	TclSetObjResultFromString(interp, "wrong # args", -1);
	return TCL_ERROR;
    }

    /*
     * THIS PROC FAILS IF AN OBJECT'S STRING REP HAS A NULL BYTE.
     */

    index = Tcl_GetStringFromObj(objv[2], &length);
    if (GetVariableIndex(interp, index, &varIndex) != TCL_OK) {
	return TCL_ERROR;
    }

    subCmd = Tcl_GetStringFromObj(objv[1], &length);
    if (strcmp(subCmd, "set") == 0) {
	if (objc != 4) {
	    goto wrongNumArgs;
	}

	/*
	 * If the object currently bound to the variable with index varIndex
	 * has ref count 1 (i.e. the object is unshared) we can modify that
	 * object directly. Otherwise, if RC>1 (i.e. the object is shared),
	 * we must create a new object to modify/set and decrement the old
	 * formerly-shared object's ref count. This is "copy on write".
	 */

	string = Tcl_GetStringFromObj(objv[3], &length);
	if ((varPtr[varIndex] != NULL) && !Tcl_IsShared(varPtr[varIndex])) {
	    Tcl_SetStringObj(varPtr[varIndex], string, length);
	} else {
	    newObjPtr = Tcl_NewStringObj(string, length);
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "get") == 0) {
	if (objc != 3) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else if (strcmp(subCmd, "append") == 0) {
	if (objc != 4) {
	    goto wrongNumArgs;
	}
	if (CheckIfVarUnset(interp, varIndex)) {
	    return TCL_ERROR;
	}
	
	/*
	 * If the object bound to variable "varIndex" is shared, we must
	 * "copy on write" and append to a copy of the object. 
	 */
	
	if (Tcl_IsShared(varPtr[varIndex])) {
	    newObjPtr = Tcl_DuplicateObj(varPtr[varIndex]);
	    SetVarToNewObj(varIndex, newObjPtr);
	}
	string = Tcl_GetStringFromObj(objv[3], &length);
	Tcl_StringObjAppend(varPtr[varIndex], string, length);
	string = Tcl_GetStringFromObj(varPtr[varIndex], &length);
	TclSetObjResultFromString(interp, string, length);
    } else {
	Tcl_Obj *resultPtr;
	
	Tcl_ResetObjResult(interp);
        resultPtr = ((Interp *) interp)->objResult;
        Tcl_StringObjAppend(resultPtr, "bad option \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[1]);
        Tcl_StringObjAppend(resultPtr, "\": must be set, get, or append", -1);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SetVarToExistingObj --
 *
 *	Utility routine to assign an existing (not new) Tcl_Obj* to a test
 *	variable. The Tcl_Obj* can be NULL.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	This routine handles ref counting details for assignment:
 *	i.e. the old value's ref count must be decremented (if not NULL) and
 *	the new one incremented (also if not NULL).
 *
 *----------------------------------------------------------------------
 */

static void
SetVarToExistingObj(varIndex, objPtr)
    int varIndex;		/* Designates the assignment variable. */
    Tcl_Obj *objPtr;		/* Existing object to assign to variable. */
{
    if (varPtr[varIndex] != NULL) {
	Tcl_DecrRefCount(varPtr[varIndex]);
    }
    varPtr[varIndex] = objPtr;
    if (objPtr != NULL) {
	Tcl_IncrRefCount(objPtr);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * SetVarToNewObj --
 *
 * 	Utility routine to assign a newly allocated Tcl_Obj* to a test
 *	variable. Newly allocated objects already have their reference
 *	count set to 1.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	This routine handles ref counting details for assignment. The
 *	ref count of the variable's old value is decremented. The new
 *	object's ref count is left unchanged at 1.
 *
 *----------------------------------------------------------------------
 */

static void
SetVarToNewObj(varIndex, newObjPtr)
    int varIndex;		/* Designates the assignment variable. */
    Tcl_Obj *newObjPtr;		/* New object to assign to the variable. */
{
    if (varPtr[varIndex] != NULL) {
	Tcl_DecrRefCount(varPtr[varIndex]);
    }
    varPtr[varIndex] = newObjPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * GetVariableIndex --
 *
 *	Utility routine to get a test variable index from the command line.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
GetVariableIndex(interp, string, indexPtr)
    Tcl_Interp *interp;         /* Interpreter for error reporting. */
    char *string;               /* String containing a variable index
				 * specified as a nonnegative number less
				 * than NUMBER_OF_OBJECT_VARS. */
    int *indexPtr;              /* Place to store converted result. */
{
    int index;
    
    if (Tcl_GetInt(interp, string, &index) != TCL_OK) {
	TclSetObjResultFromStrResult(interp);
	return TCL_ERROR;
    }
    if (index < 0 || index >= NUMBER_OF_OBJECT_VARS) {
	TclSetObjResultFromString(interp, "bad variable index", -1);
	return TCL_ERROR;
    }

    *indexPtr = index;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CheckIfVarUnset --
 *
 *	Utility procedure that checks whether a test variable is readable:
 *	i.e., that varPtr[varIndex] is non-NULL.
 *
 * Results:
 *	1 if the test variable is unset (NULL); 0 otherwise.
 *
 * Side effects:
 *	Sets interp->objResult to an error message if the variable is
 *	unset (NULL).
 *
 *----------------------------------------------------------------------
 */

static int
CheckIfVarUnset(interp, varIndex)
    Tcl_Interp *interp;		/* Interpreter for error reporting. */
    int varIndex;		/* Index of the test variable to check. */
{
    if (varPtr[varIndex] == NULL) {
	char buf[100];
	sprintf(buf, "variable %d is unset (NULL)", varIndex);
	TclSetObjResultFromString(interp, buf, -1);
	return 1;
    }
    return 0;
}
